% :- type row(T)    == list(T).
% :- type matrix(T) == list(row(T)).
% :- type field     == int.
% :- type dir       ---> n                % north
%                   ;    e                % east
%                   ;    s                % south
%                   ;    w                % west
%                   ;    ne               % northeast
%                   ;    nw               % northwest
%                   ;    se               % southeast
%                   ;    sw.              % southwest

:- module(cikcakk_frame, [
	cikcakk_in/3,   % cikcakk_in(file::in, int::out, matrix(field)::out)
	cikcakk_out/3,	% cikcakk_out(file::in, matrix(field)::in,
		 	%                       matrix(dir)::in)
	solve/2,	% solve(file::in, file::in)
	time/2		% time(file::in, file::in)
			]
	 ).

:- use_module(library(lists), [append/3, nth/3, max_list/2]).

%-------------- Bejaratok ------------------------------------

% cikcakk_in(+File, -Max, -Table):
% A File-bol beolvashato egy feladat, melyet a Table matrix ir le, melyben
% a maximalis ertek Max.
cikcakk_in(File, Max, Table) :-
	open_file(File, read, Stream),
	read_in(Stream, Characters),
	close(Stream),
	parse_data(Characters, Max, Table).

% cikcakk_out(+File, +Table, +Solution):
% A Table feladat Solution megoldasat kiirja a File allomanyba.
cikcakk_out(File, Table, Solution) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Table, Solution),
		     (set_output(OOut),close(Stream))).

% solve(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
solve(FileIn, FileOut) :-
	cikcakk_in(FileIn, Max, Table),
	ensure_loaded(user:cikcakk),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Table, Max),
		     (set_output(OOut),close(Stream))).

% print_solutions(+Table, +Max):
% A Max maximalis mezo-erteku Table feladat minden megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Table, Max) :-
	user:cikcakk(Table, Max, Solution),
	pretty_print(Table, Solution),
	fail.
print_solutions(_, _).

% time(+FileIn, +FileOut):
% Mint solve/2, de a futashoz szukseges idot is kiirja.
time(FileIn, FileOut) :-
	statistics(runtime, [T0|_]),
	solve(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Table: ~a, time: ~3d sec~n', [FileIn,T]).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Max, -Table):
% Chars fuzerbol kielemezheto egy feladat, amit a Table matrix ir le, melyben
% a maximalis ertek Max.
parse_data(Chars, Max, Table) :-
	first_int(Max, Chars, Chars1),
	int_lists(Table, Chars1, Rest),
	correct_syntax(Rest),
	correct_matrix(Table),
	matrix_max(Table, Max).

%-------------- Ellenorzesek ---------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :- 
	Rest = [_|_], !, 
	chars_of_type(mid_line, L, Rest, _),
	warning('non-digit character in line: "~s"', [L]),
	fail.
correct_syntax(_).

% correct_matrix(+M):
% M valoban egy matrix, azaz minden sora egyforma hosszu.
correct_matrix([Row1|Rows]) :-
	length(Row1, Len1),
	nth(I, Rows, Row),
	length(Row, Len),
	Len =\= Len1, !,
	I1 is I+1,
	warning(
	  'non-rectangular: line 1 length: ~d, line ~d length: ~d',
	  [Len1, I1, Len]),
	fail.
correct_matrix(_).

%-------------- DCG elemzes ----------------------------------

% first_int(-Int) -->
% Kielemezheto az Int egesz, amit esetleg nem lathato karakterek
% eloznek meg.
first_int(Int) -->
	chars_of_type(layout, _),
	int_number(Int).

% int_lists(-LL) -->
% Kielemezheto az LL egeszek listainak listaja.
int_lists([IntList|IntLists]) -->
	chars_of_type(layout, _),
	int_list(IntList), !, 
	int_lists(IntLists).
int_lists([]) --> [].

% int_list(-IntList) -->
% Kielemezheto az IntList egeszek listaja, ujsorral lezarva.  
% A bemenet nem kezdodhet nem lathato karakterrel.
int_list([Int|IntList]) -->
	int_number(Int), !,
	chars_of_type(horiz_sp, _),
	int_list(IntList).
int_list([]) -->
	char_of_type(vert_sp, _).

% int_number(-Int) -->
% Kielemezheto egy (tizes szamrendszerbeli) Int nem-negativ egesz szam.
int_number(Int) -->
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% layout_char(Dir, C):
% C egy Dir (horiz_sp|vert_sp) iranyu nem lathato karakter.
layout_char(horiz_sp, 0' ).
layout_char(horiz_sp, 0'\t).
layout_char(vert_sp,  0'\n).
layout_char(vert_sp,  0'\r).
layout_char(vert_sp,  0'\f).
layout_char(vert_sp,  0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	layout_char(_, C).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ layout_char(vert_sp, C).
char_type(Dir, C) :-
	layout_char(Dir, C).

%-------------- Kiiratas -------------------------------------

% pretty_print(+Table, +Solution):
% kiirja a Table feladat  Solution megoldasat.
pretty_print(Table, Solution) :-
	Table = [FRow|_],
	length(FRow, NumCols),
	matrix_max(Table, N),
	Pad is integer(log(10, max(N,1)))+3,
	print_separator(NumCols, Pad),
	pretty_print_rows(Table, Solution, Pad, NumCols), nl.

% pretty_print_rows(+Rs, +Ss, +Pad, +NumCols):
% Az Rs feladat- es Ss megoldassorokat kiirja, a szamokat egy Pad
% szelessegu mezoben jobbra igazitva elejuk irva az iranyt.  Az egyes
% sorokat NumCols oszlopnyi szeparator hatarolja.
pretty_print_rows([], [], _Pad, _NumCols).
pretty_print_rows([R|Rs], [S|Ss], Pad, NumCols) :-
	put_code(0'|), pretty_print_row(R, S, Pad),
	print_separator(NumCols, Pad),
	pretty_print_rows(Rs, Ss, Pad, NumCols).

% pretty_print_row(+R, +S, +Pad):
% Az R feladat- es az S megoldassort (a szamokat es iranyokat egy min. Pad
% szelessegu mezoben jobbra igazitva) kiirja.
pretty_print_row([], [], _) :-
	nl.
pretty_print_row([V|Vs], Ds, Pad) :-
	(   Ds == [] -> D = '.', Ds1 = []
	;   Ds = [D|Ds1]
	),
	format('~|~t~a~d~*+|', [D,V,Pad]),
	pretty_print_row(Vs, Ds1, Pad).

% print_separator(+N, +P):
% Egy N oszlopos, P mezoszelessegu szeparator-sort nyomtat.
print_separator(0, _) :-
	write('+\n').
print_separator(N, P) :-
	N > 0,
	format('+~*c', [P,0'-]),
	N1 is N-1,
	print_separator(N1, P).

%-------------- Seged-eljarasok ------------------------------

% append_lists(+Ls, ?L):
% Az Ls listak osszefuzesevel kapott lista L.
append_lists([], []).
append_lists([H|T], C) :-
	append(H, R, C),
	append_lists(T, R).

% matrix_max(+LLs, ?Max):
% Az LLs matrix elemeinek maximuma Max.
matrix_max(LLs, Max) :-
	append_lists(LLs, All),
	max_list(All, Max).

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
